import numpy as np
import plotly.graph_objects as go
from finalysis.instruments import Instrument

def payoff(
  strategy: Instrument,
  price: float,
  *,
  min_price: float | None = None,
  max_price: float | None = None,
  steps: int = 1000,
  relative: bool = False,
  show_current_price: bool = True,
) -> go.Figure:
  """
  Generate an interactive payoff plot for a strategy.

  Parameters:
  - strategy: an Instrument implementing .payoff
  - price: current price of the underlying
  - min_price, max_price: price range for the plot (optional)
  - steps: number of points in the plot
  - relative: show PnL as a percentage of current price
  - show_current_price: add a vertical line for current price

  Returns:
  - plotly.graph_objects.Figure
  """
  X = np.linspace(min_price or 0.75 * price, max_price or 1.5 * price, steps)
  Y = strategy.payoff({'final_price': X})

  if relative:
      Y = Y / price

  fig = go.Figure()

  fig.add_trace(go.Scatter(
      x=X, y=Y,
      mode='lines',
      name='PnL',
      hovertemplate='Final Price: %{x:.2f}<br>PnL: %{y:.2f}<extra></extra>'
  ))

  if show_current_price:
      fig.add_vline(
          x=price,
          line=dict(color='red', dash='dash'),
          annotation_text="Current Price",
          annotation_position="top right"
      )

  fig.update_layout(
      title="Strategy Payoff",
      xaxis_title="Final Price",
      yaxis_title="PnL" if not relative else "Relative PnL",
      hovermode="x unified",
      showlegend=True,
      template="plotly_white"
  )

  return fig



from dataclasses import dataclass
import numpy as np
from .instrument import Instrument

@dataclass
class SellHighDI(Instrument):
  """Sell High Dual Investment"""
  rate: float
  """interest rate (e.g 0.05)"""
  strike: float
  price: float
  def payoff(self, data):
    rel = np.minimum(data['final_price'], self.strike)*(1+self.rate) - self.price
    return rel * self.quantity
  
@dataclass
class BuyLowDI(Instrument):
  """Buy Low Dual Investment"""
  rate: float
  """interest rate (e.g 0.05)"""
  strike: float
  price: float

  def payoff(self, data):
    rel = np.where(
      data['final_price'] < self.strike,
      data['final_price']*(1+self.rate) - self.strike,
      self.rate*self.price
    )
    return rel * self.quantity
  
@dataclass
class SmartLeverage(Instrument):
  """https://www.bybit.com/en/earn/smart-leverage"""
  leverage: float
  breakeven: float

  def payoff(self, data):
    gain = self.leverage * (data['final_price'] - self.breakeven) / self.breakeven
    return np.maximum(gain, -self.quantity)


from .instrument import Instrument, Data, Sum
from .basics import Spot, Option, Future
from .spot import SpotLimit, SpotGrid
from .structured import BuyLowDI, SellHighDI, SmartLeverage

__all__ = [
  'Instrument', 'Data', 'Sum',
  'Spot', 'Option', 'Future',
  'SpotLimit', 'SpotGrid',
  'BuyLowDI', 'SellHighDI', 'SmartLeverage',
]


from typing_extensions import Literal
from dataclasses import dataclass
import numpy as np
from .instrument import Instrument

@dataclass
class Spot(Instrument):
  price: float
  def payoff(self, data):
    return (data['final_price'] - self.price) * self.quantity

@dataclass
class Option(Instrument):
  strike: float
  premium: float
  kind: Literal['call', 'put']

  def payoff(self, data):
    if self.kind == 'put':
      diff = self.strike - data['final_price']
    else:
      diff = data['final_price'] - self.strike
    rel = np.maximum(diff, 0) - self.premium
    return rel * self.quantity
  
@dataclass
class Future(Instrument):
  price: float
  kind: Literal['long', 'short']
  
  def payoff(self, data):
    sign = 1 if self.kind == 'long' else -1
    return (data['final_price'] - self.price) * self.quantity * sign
  


from typing_extensions import TypeVar, Generic, TypedDict
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, replace
import numpy as np

class Data(TypedDict):
  final_price: np.ndarray | float

D = TypeVar('D', bound=Data, default=Data, contravariant=True)
D2 = TypeVar('D2', bound=Data, default=Data, contravariant=True)

@dataclass
class Instrument(ABC, Generic[D]):
  quantity: float = field(default=1, kw_only=True)
  @abstractmethod
  def payoff(self, data: D, /) -> np.ndarray:
    ...

  def __add__(self, other: 'Instrument[D2]') -> 'Sum[D2]':
    return Sum([self, other])
  
  def __mul__(self, quantity: float) -> 'Instrument[D]':
    return replace(self, quantity=quantity)
  
  __rmul__ = __mul__

  def __neg__(self) -> 'Instrument[D]':
    return self * -1
  
  def __sub__(self, other: 'Instrument[D2]') -> 'Sum[D2]':
    return self + -other
  
@dataclass
class Sum(Instrument[D]):
  instruments: list[Instrument[D]]
  def payoff(self, data: D) -> np.ndarray:
    return np.sum([instrument.payoff(data) for instrument in self.instruments], axis=0)
  
  def __add__(self, other: 'Instrument[D2]') -> 'Sum[D2]':
    return replace(self, instruments=self.instruments + [other])
  
  def __mul__(self, quantity: float) -> 'Sum[D]':
    return replace(self, instruments=[instrument * quantity for instrument in self.instruments])


from typing import Sequence, TypedDict
from dataclasses import dataclass
import numpy as np
from .instrument import Instrument, Data

class RangeData(Data):
  min_price: float

@dataclass
class SpotLimit(Instrument[RangeData]):
  price: float
  
  def payoff(self, data):
    if data['min_price'] <= self.price:
      return (data['final_price'] - self.price) * self.quantity
    else:
      return np.array(0.)
    
class Candle(TypedDict):
  open: float
  high: float
  low: float
  close: float

class GridData(Data):
  candles: Sequence[Candle]
    
@dataclass
class SpotGrid(Instrument[GridData]):
  buy_price: float
  sell_price: float
  fee: float = 0
  debug: bool = False

  def payoff(self, data: GridData):
    import numpy as np
    payoff = np.zeros_like(data['final_price'])
    bought = False

    def step(price: float) -> bool:
      nonlocal payoff, bought
      if not bought and price <= self.buy_price:
        payoff -= self.buy_price*(1+self.fee)
        if self.debug:
          print('Buying at:', price, 'payoff:', payoff)
        bought = True
        stepped = True
      
      elif bought and price >= self.sell_price:
        payoff += self.sell_price*(1-self.fee)
        if self.debug:
          print('Selling at:', price, 'payoff:', payoff)
        bought = False
        stepped = True

      else:
        stepped = False
      
      return stepped


    for c in data['candles']:
      step(c['open'])
      step(c['high']) or step(c['low']) # type: ignore
      step(c['close'])
      
    if bought:
      payoff += data['final_price']
      if self.debug:
        print(f"Final value: {data['final_price']}, payoff:", payoff)

    return payoff * self.quantity